@@ -14,10 +14,12 @@
 
 """A library of random helper functionality."""
 from pathlib import Path
+import argparse
 import enum
 import sys
 import stat
 import time
+import abc
 import platform, subprocess, operator, os, shlex, shutil, re
 import collections
 from functools import lru_cache, wraps, total_ordering
@@ -26,10 +28,12 @@ from tempfile import TemporaryDirectory
 import typing as T
 import uuid
 import textwrap
+import copy
 
 from mesonbuild import mlog
 
 if T.TYPE_CHECKING:
+    from .._typing import ImmutableListProtocol
     from ..build import ConfigurationData
     from ..coredata import KeyedOptionDictType, UserOption
     from ..compilers.compilers import CompilerType
@@ -45,19 +49,21 @@ __all__ = [
     'an_unpicklable_object',
     'python_command',
     'project_meson_versions',
+    'HoldableObject',
+    'SecondLevelHolder',
     'File',
     'FileMode',
     'GitException',
     'LibType',
     'MachineChoice',
     'MesonException',
+    'MesonBugException',
     'EnvironmentException',
     'FileOrString',
     'GitException',
     'OptionKey',
     'dump_conf_header',
     'OptionOverrideProxy',
-    'OptionProxy',
     'OptionType',
     'OrderedSet',
     'PerMachine',
@@ -65,6 +71,7 @@ __all__ = [
     'PerThreeMachine',
     'PerThreeMachineDefaultable',
     'ProgressBar',
+    'RealPathAction',
     'TemporaryDirectoryWinProof',
     'Version',
     'check_direntry_issues',
@@ -107,6 +114,7 @@ __all__ = [
     'is_qnx',
     'is_sunos',
     'is_windows',
+    'is_wsl',
     'iter_regexin_iter',
     'join_args',
     'listify',
@@ -126,11 +134,11 @@ __all__ = [
     'substitute_values',
     'substring_is_in_list',
     'typeslistify',
-    'unholder',
     'verbose_git',
     'version_compare',
     'version_compare_condition_with_min',
     'version_compare_many',
+    'search_version',
     'windows_proof_rm',
     'windows_proof_rmtree',
 ]
@@ -162,6 +170,14 @@ class MesonException(Exception):
         self.colno = colno
 
 
+class MesonBugException(MesonException):
+    '''Exceptions thrown when there is a clear Meson bug that should be reported'''
+
+    def __init__(self, msg: str, file: T.Optional[str] = None,
+                 lineno: T.Optional[int] = None, colno: T.Optional[int] = None):
+        super().__init__(msg + '\n\n    This is a Meson bug and should be reported!',
+                         file=file, lineno=lineno, colno=colno)
+
 class EnvironmentException(MesonException):
     '''Exceptions thrown while processing and creating the build environment'''
 
@@ -213,7 +229,7 @@ def set_meson_command(mainfile: str) -> None:
         _meson_command = python_command + [mainfile]
     # We print this value for unit tests.
     if 'MESON_COMMAND_TESTS' in os.environ:
-        mlog.log('meson_command is {!r}'.format(_meson_command))
+        mlog.log(f'meson_command is {_meson_command!r}')
 
 
 def get_meson_command() -> T.Optional[T.List[str]]:
@@ -244,11 +260,11 @@ def check_direntry_issues(direntry_array: T.Union[T.List[T.Union[str, bytes]], s
         for de in direntry_array:
             if is_ascii_string(de):
                 continue
-            mlog.warning(textwrap.dedent('''
-                You are using {!r} which is not a Unicode-compatible
-                locale but you are trying to access a file system entry called {!r} which is
+            mlog.warning(textwrap.dedent(f'''
+                You are using {e!r} which is not a Unicode-compatible
+                locale but you are trying to access a file system entry called {de!r} which is
                 not pure ASCII. This may cause problems.
-                '''.format(e, de)), file=sys.stderr)
+                '''), file=sys.stderr)
 
 
 # Put this in objects that should not get dumped to pickle files
@@ -256,6 +272,18 @@ def check_direntry_issues(direntry_array: T.Union[T.List[T.Union[str, bytes]], s
 import threading
 an_unpicklable_object = threading.Lock()
 
+class HoldableObject(metaclass=abc.ABCMeta):
+    ''' Dummy base class for all objects that can be
+        held by an interpreter.baseobjects.ObjectHolder '''
+
+class SecondLevelHolder(HoldableObject, metaclass=abc.ABCMeta):
+    ''' A second level object holder. The primary purpose
+        of such objects is to hold multiple objects with one
+        default option. '''
+
+    @abc.abstractmethod
+    def get_default_object(self) -> HoldableObject: ...
+
 class FileMode:
     # The first triad is for owner permissions, the second for group permissions,
     # and the third for others (everyone else).
@@ -282,8 +310,8 @@ class FileMode:
                                       '[r-][w-][xsS-]' # Group perms
                                       '[r-][w-][xtT-]') # Others perms
 
-    def __init__(self, perms: T.Optional[str] = None, owner: T.Optional[str] = None,
-                 group: T.Optional[str] = None):
+    def __init__(self, perms: T.Optional[str] = None, owner: T.Union[str, int, None] = None,
+                 group: T.Union[str, int, None] = None):
         self.perms_s = perms
         self.perms = self.perms_s_to_bits(perms)
         self.owner = owner
@@ -304,11 +332,9 @@ class FileMode:
             return -1
         eg = 'rwxr-xr-x'
         if not isinstance(perms_s, str):
-            msg = 'Install perms must be a string. For example, {!r}'
-            raise MesonException(msg.format(eg))
+            raise MesonException(f'Install perms must be a string. For example, {eg!r}')
         if len(perms_s) != 9 or not cls.symbolic_perms_regex.match(perms_s):
-            msg = 'File perms {!r} must be exactly 9 chars. For example, {!r}'
-            raise MesonException(msg.format(perms_s, eg))
+            raise MesonException(f'File perms {perms_s!r} must be exactly 9 chars. For example, {eg!r}')
         perms = 0
         # Owner perms
         if perms_s[0] == 'r':
@@ -349,11 +375,13 @@ class FileMode:
         return perms
 
 dot_C_dot_H_warning = """You are using .C or .H files in your project. This is deprecated.
-         Currently, Meson treats this as C code, but this
-            might change in the future, breaking your build.
-         You code also might be already broken on gcc and clang.
-         See https://github.com/mesonbuild/meson/pull/8239 for the discussions."""
-class File:
+         Currently, Meson treats this as C++ code, but they
+            used to be treated as C code.
+         Note that the situation is a bit more complex if you are using the
+         Visual Studio compiler, as it treats .C files as C code, unless you add
+         the /TP compiler flag, but this is unreliable.
+         See https://github.com/mesonbuild/meson/pull/8747 for the discussions."""
+class File(HoldableObject):
     def __init__(self, is_built: bool, subdir: str, fname: str):
         if fname.endswith(".C") or fname.endswith(".H"):
             mlog.warning(dot_C_dot_H_warning, once=True)
@@ -404,8 +432,11 @@ class File:
     def endswith(self, ending: str) -> bool:
         return self.fname.endswith(ending)
 
-    def split(self, s: str) -> T.List[str]:
-        return self.fname.split(s)
+    def split(self, s: str, maxsplit: int = -1) -> T.List[str]:
+        return self.fname.split(s, maxsplit=maxsplit)
+
+    def rsplit(self, s: str, maxsplit: int = -1) -> T.List[str]:
+        return self.fname.rsplit(s, maxsplit=maxsplit)
 
     def __eq__(self, other: object) -> bool:
         if not isinstance(other, File):
@@ -427,7 +458,7 @@ def get_compiler_for_source(compilers: T.Iterable['CompilerType'], src: str) ->
     for comp in compilers:
         if comp.can_compile(src):
             return comp
-    raise MesonException('No specified compiler can handle file {!s}'.format(src))
+    raise MesonException(f'No specified compiler can handle file {src!s}')
 
 
 def classify_unity_sources(compilers: T.Iterable['CompilerType'], sources: T.Iterable[str]) -> T.Dict['CompilerType', T.List[str]]:
@@ -486,7 +517,7 @@ class PerMachine(T.Generic[_T]):
         return unfreeze
 
     def __repr__(self) -> str:
-        return 'PerMachine({!r}, {!r})'.format(self.build, self.host)
+        return f'PerMachine({self.build!r}, {self.host!r})'
 
 
 class PerThreeMachine(PerMachine[_T]):
@@ -521,16 +552,16 @@ class PerThreeMachine(PerMachine[_T]):
         return self.build == self[machine]
 
     def __repr__(self) -> str:
-        return 'PerThreeMachine({!r}, {!r}, {!r})'.format(self.build, self.host, self.target)
+        return f'PerThreeMachine({self.build!r}, {self.host!r}, {self.target!r})'
 
 
 class PerMachineDefaultable(PerMachine[T.Optional[_T]]):
     """Extends `PerMachine` with the ability to default from `None`s.
     """
-    def __init__(self) -> None:
-        super().__init__(None, None)
+    def __init__(self, build: T.Optional[_T] = None, host: T.Optional[_T] = None) -> None:
+        super().__init__(build, host)
 
-    def default_missing(self) -> "PerMachine[T.Optional[_T]]":
+    def default_missing(self) -> "PerMachine[_T]":
         """Default host to build
 
         This allows just specifying nothing in the native case, and just host in the
@@ -542,7 +573,22 @@ class PerMachineDefaultable(PerMachine[T.Optional[_T]]):
         return freeze
 
     def __repr__(self) -> str:
-        return 'PerMachineDefaultable({!r}, {!r})'.format(self.build, self.host)
+        return f'PerMachineDefaultable({self.build!r}, {self.host!r})'
+
+    @classmethod
+    def default(cls, is_cross: bool, build: _T, host: _T) -> PerMachine[_T]:
+        """Easy way to get a defaulted value
+
+        This allows simplifying the case where you can control whether host and
+        build are separate or not with a boolean. If the is_cross value is set
+        to true then the optional host value will be used, otherwise the host
+        will be set to the build value.
+        """
+        m = cls(build)
+        if is_cross:
+            m.host = host
+        return m.default_missing()
+
 
 
 class PerThreeMachineDefaultable(PerMachineDefaultable, PerThreeMachine[T.Optional[_T]]):
@@ -566,7 +612,7 @@ class PerThreeMachineDefaultable(PerMachineDefaultable, PerThreeMachine[T.Option
         return freeze
 
     def __repr__(self) -> str:
-        return 'PerThreeMachineDefaultable({!r}, {!r}, {!r})'.format(self.build, self.host, self.target)
+        return f'PerThreeMachineDefaultable({self.build!r}, {self.host!r}, {self.target!r})'
 
 
 def is_sunos() -> bool:
@@ -597,6 +643,8 @@ def is_windows() -> bool:
     platname = platform.system().lower()
     return platname == 'windows'
 
+def is_wsl() -> bool:
+    return is_linux() and 'microsoft' in platform.release().lower()
 
 def is_cygwin() -> bool:
     return sys.platform == 'cygwin'
@@ -639,7 +687,7 @@ def exe_exists(arglist: T.List[str]) -> bool:
 
 
 @lru_cache(maxsize=None)
-def darwin_get_object_archs(objpath: str) -> T.List[str]:
+def darwin_get_object_archs(objpath: str) -> 'ImmutableListProtocol[str]':
     '''
     For a specific object (executable, static library, dylib, etc), run `lipo`
     to fetch the list of archs supported by it. Supports both thin objects and
@@ -647,7 +695,7 @@ def darwin_get_object_archs(objpath: str) -> T.List[str]:
     '''
     _, stdo, stderr = Popen_safe(['lipo', '-info', objpath])
     if not stdo:
-        mlog.debug('lipo {}: {}'.format(objpath, stderr))
+        mlog.debug(f'lipo {objpath}: {stderr}')
         return None
     stdo = stdo.rsplit(': ', 1)[1]
     # Convert from lipo-style archs to meson-style CPUs
@@ -682,6 +730,12 @@ def detect_vcs(source_dir: T.Union[str, Path]) -> T.Optional[T.Dict[str, str]]:
 
 def current_vs_supports_modules() -> bool:
     vsver = os.environ.get('VSCMD_VER', '')
+    nums = vsver.split('.', 2)
+    major = int(nums[0])
+    if major >= 17:
+        return True
+    if major == 16 and int(nums[1]) >= 10:
+        return True
     return vsver.startswith('16.9.0') and '-pre.' in vsver
 
 # a helper class which implements the same version ordering as RPM
@@ -701,10 +755,10 @@ class Version:
         self._v = sequences3
 
     def __str__(self) -> str:
-        return '%s (V=%s)' % (self._s, str(self._v))
+        return '{} (V={})'.format(self._s, str(self._v))
 
     def __repr__(self) -> str:
-        return '<Version: {}>'.format(self._s)
+        return f'<Version: {self._s}>'
 
     def __lt__(self, other: object) -> bool:
         if isinstance(other, Version):
@@ -839,6 +893,49 @@ def version_compare_condition_with_min(condition: str, minimum: str) -> bool:
 
     return T.cast(bool, cmpop(Version(minimum), Version(condition)))
 
+def search_version(text: str) -> str:
+    # Usually of the type 4.1.4 but compiler output may contain
+    # stuff like this:
+    # (Sourcery CodeBench Lite 2014.05-29) 4.8.3 20140320 (prerelease)
+    # Limiting major version number to two digits seems to work
+    # thus far. When we get to GCC 100, this will break, but
+    # if we are still relevant when that happens, it can be
+    # considered an achievement in itself.
+    #
+    # This regex is reaching magic levels. If it ever needs
+    # to be updated, do not complexify but convert to something
+    # saner instead.
+    # We'll demystify it a bit with a verbose definition.
+    version_regex = re.compile(r"""
+    (?<!                # Zero-width negative lookbehind assertion
+        (
+            \d          # One digit
+            | \.        # Or one period
+        )               # One occurrence
+    )
+    # Following pattern must not follow a digit or period
+    (
+        \d{1,2}         # One or two digits
+        (
+            \.\d+       # Period and one or more digits
+        )+              # One or more occurrences
+        (
+            -[a-zA-Z0-9]+   # Hyphen and one or more alphanumeric
+        )?              # Zero or one occurrence
+    )                   # One occurrence
+    """, re.VERBOSE)
+    match = version_regex.search(text)
+    if match:
+        return match.group(0)
+
+    # try a simpler regex that has like "blah 2020.01.100 foo" or "blah 2020.01 foo"
+    version_regex = re.compile(r"(\d{1,4}\.\d{1,4}\.?\d{0,4})")
+    match = version_regex.search(text)
+    if match:
+        return match.group(0)
+
+    return 'unknown version'
+
 
 def default_libdir() -> str:
     if is_debianlike():
@@ -929,7 +1026,7 @@ if is_windows():
     # https://blogs.msdn.microsoft.com/twistylittlepassagesallalike/2011/04/23/everyone-quotes-command-line-arguments-the-wrong-way/
 
     _whitespace = ' \t\n\r'
-    _find_unsafe_char = re.compile(r'[{}"]'.format(_whitespace)).search
+    _find_unsafe_char = re.compile(fr'[{_whitespace}"]').search
 
     def quote_arg(arg: str) -> str:
         if arg and not _find_unsafe_char(arg):
@@ -1029,9 +1126,9 @@ def do_replacement(regex: T.Pattern[str], line: str, variable_format: str,
                 elif isinstance(var, int):
                     var_str = str(var)
                 else:
-                    msg = 'Tried to replace variable {!r} value with ' \
-                          'something other than a string or int: {!r}'
-                    raise MesonException(msg.format(varname, var))
+                    msg = f'Tried to replace variable {varname!r} value with ' \
+                          f'something other than a string or int: {var!r}'
+                    raise MesonException(msg)
             else:
                 missing_variables.add(varname)
             return var_str
@@ -1070,7 +1167,7 @@ def do_define(regex: T.Pattern[str], line: str, confdata: 'ConfigurationData', v
             result = v
         else:
             result = get_cmake_define(line, confdata)
-        result = '#define %s %s\n' % (varname, result)
+        result = f'#define {varname} {result}\n'
         (result, missing_variable) = do_replacement(regex, result, variable_format, confdata)
         return result
     else:
@@ -1084,10 +1181,10 @@ def get_variable_regex(variable_format: str = 'meson') -> T.Pattern[str]:
     elif variable_format == 'cmake':
         regex = re.compile(r'(?:\\\\)+(?=\\?\$)|\\\${|\${([-a-zA-Z0-9_]+)}')
     else:
-        raise MesonException('Format "{}" not handled'.format(variable_format))
+        raise MesonException(f'Format "{variable_format}" not handled')
     return regex
 
-def do_conf_str (data: list, confdata: 'ConfigurationData', variable_format: str,
+def do_conf_str (src: str, data: list, confdata: 'ConfigurationData', variable_format: str,
                  encoding: str = 'utf-8') -> T.Tuple[T.List[str],T.Set[str], bool]:
     def line_is_valid(line : str, variable_format: str) -> bool:
         if variable_format == 'meson':
@@ -1115,7 +1212,7 @@ def do_conf_str (data: list, confdata: 'ConfigurationData', variable_format: str
             line = do_define(regex, line, confdata, variable_format)
         else:
             if not line_is_valid(line,variable_format):
-                raise MesonException('Format "{}" mismatched'.format(variable_format))
+                raise MesonException(f'Format error in {src}: saw "{line.strip()}" when format set to "{variable_format}"')
             line, missing = do_replacement(regex, line, variable_format, confdata)
             missing_variables.update(missing)
             if missing:
@@ -1130,15 +1227,15 @@ def do_conf_file(src: str, dst: str, confdata: 'ConfigurationData', variable_for
         with open(src, encoding=encoding, newline='') as f:
             data = f.readlines()
     except Exception as e:
-        raise MesonException('Could not read input file %s: %s' % (src, str(e)))
+        raise MesonException(f'Could not read input file {src}: {e!s}')
 
-    (result, missing_variables, confdata_useless) = do_conf_str(data, confdata, variable_format, encoding)
+    (result, missing_variables, confdata_useless) = do_conf_str(src, data, confdata, variable_format, encoding)
     dst_tmp = dst + '~'
     try:
         with open(dst_tmp, 'w', encoding=encoding, newline='') as f:
             f.writelines(result)
     except Exception as e:
-        raise MesonException('Could not write output file %s: %s' % (dst, str(e)))
+        raise MesonException(f'Could not write output file {dst}: {e!s}')
     shutil.copymode(src, dst_tmp)
     replace_if_different(dst, dst_tmp)
     return missing_variables, confdata_useless
@@ -1178,11 +1275,11 @@ def dump_conf_header(ofilename: str, cdata: 'ConfigurationData', output_format:
                         ofile.write('; %s\n' % line)
             if isinstance(v, bool):
                 if v:
-                    ofile.write('%sdefine %s\n\n' % (prefix, k))
+                    ofile.write(f'{prefix}define {k}\n\n')
                 else:
-                    ofile.write('%sundef %s\n\n' % (prefix, k))
+                    ofile.write(f'{prefix}undef {k}\n\n')
             elif isinstance(v, (int, str)):
-                ofile.write('%sdefine %s %s\n\n' % (prefix, k, v))
+                ofile.write(f'{prefix}define {k} {v}\n\n')
             else:
                 raise MesonException('Unknown data type in configuration file entry: ' + k)
     replace_if_different(ofilename, ofilename_tmp)
@@ -1204,26 +1301,6 @@ def replace_if_different(dst: str, dst_tmp: str) -> None:
         os.unlink(dst_tmp)
 
 
-@T.overload
-def unholder(item: 'ObjectHolder[_T]') -> _T: ...
-
-@T.overload
-def unholder(item: T.List['ObjectHolder[_T]']) -> T.List[_T]: ...
-
-@T.overload
-def unholder(item: T.List[_T]) -> T.List[_T]: ...
-
-@T.overload
-def unholder(item: T.List[T.Union[_T, 'ObjectHolder[_T]']]) -> T.List[_T]: ...
-
-def unholder(item):  # type: ignore  # TODO fix overload (somehow)
-    """Get the held item of an object holder or list of object holders."""
-    if isinstance(item, list):
-        return [i.held_object if hasattr(i, 'held_object') else i for i in item]
-    if hasattr(item, 'held_object'):
-        return item.held_object
-    return item
-
 
 def listify(item: T.Any, flatten: bool = True) -> T.List[T.Any]:
     '''
@@ -1282,7 +1359,7 @@ def expand_arguments(args: T.Iterable[str]) -> T.Optional[T.List[str]]:
 
         args_file = arg[1:]
         try:
-            with open(args_file) as f:
+            with open(args_file, encoding='utf-8') as f:
                 extended_args = f.read().split()
             expended_args += extended_args
         except Exception as e:
@@ -1292,7 +1369,7 @@ def expand_arguments(args: T.Iterable[str]) -> T.Optional[T.List[str]]:
     return expended_args
 
 
-def partition(pred: T.Callable[[_T], object], iterable: T.Iterator[_T]) -> T.Tuple[T.Iterator[_T], T.Iterator[_T]]:
+def partition(pred: T.Callable[[_T], object], iterable: T.Iterable[_T]) -> T.Tuple[T.Iterator[_T], T.Iterator[_T]]:
     """Use a predicate to partition entries into false entries and true
     entries.
 
@@ -1374,15 +1451,14 @@ def _substitute_values_check_errors(command: T.List[str], values: T.Dict[str, st
         # Error out if any input-derived templates are present in the command
         match = iter_regexin_iter(inregex, command)
         if match:
-            m = 'Command cannot have {!r}, since no input files were specified'
-            raise MesonException(m.format(match))
+            raise MesonException(f'Command cannot have {match!r}, since no input files were specified')
     else:
         if len(values['@INPUT@']) > 1:
             # Error out if @PLAINNAME@ or @BASENAME@ is present in the command
             match = iter_regexin_iter(inregex[1:], command)
             if match:
-                raise MesonException('Command cannot have {!r} when there is '
-                                     'more than one input file'.format(match))
+                raise MesonException(f'Command cannot have {match!r} when there is '
+                                     'more than one input file')
         # Error out if an invalid @INPUTnn@ template was specified
         for each in command:
             if not isinstance(each, str):
@@ -1486,7 +1562,7 @@ def get_filenames_templates_dict(inputs: T.List[str], outputs: T.List[str]) -> T
         values['@INPUT@'] = inputs
         for (ii, vv) in enumerate(inputs):
             # Write out @INPUT0@, @INPUT1@, ...
-            values['@INPUT{}@'.format(ii)] = vv
+            values[f'@INPUT{ii}@'] = vv
         if len(inputs) == 1:
             # Just one value, substitute @PLAINNAME@ and @BASENAME@
             values['@PLAINNAME@'] = plain = os.path.basename(inputs[0])
@@ -1495,7 +1571,7 @@ def get_filenames_templates_dict(inputs: T.List[str], outputs: T.List[str]) -> T
         # Gather values derived from the outputs, similar to above.
         values['@OUTPUT@'] = outputs
         for (ii, vv) in enumerate(outputs):
-            values['@OUTPUT{}@'.format(ii)] = vv
+            values[f'@OUTPUT{ii}@'] = vv
         # Outdir should be the same for all outputs
         values['@OUTDIR@'] = os.path.dirname(outputs[0])
         # Many external programs fail on empty arguments.
@@ -1521,9 +1597,16 @@ def windows_proof_rmtree(f: str) -> None:
     # be scanning files you are trying to delete. The only
     # way to fix this is to try again and again.
     delays = [0.1, 0.1, 0.2, 0.2, 0.2, 0.5, 0.5, 1, 1, 1, 1, 2]
-    # Start by making the tree wriable.
-    _make_tree_writable(f)
+    writable = False
     for d in delays:
+        try:
+            # Start by making the tree writable.
+            if not writable:
+                _make_tree_writable(f)
+                writable = True
+        except PermissionError:
+            time.sleep(d)
+            continue
         try:
             shutil.rmtree(f)
             return
@@ -1716,7 +1799,7 @@ class ProgressBarFallback:  # lgtm [py/iter-returns-non-self]
         if self.total and bar_type == 'download':
             print('Download size:', self.total)
         if desc:
-            print('{}: '.format(desc), end='')
+            print(f'{desc}: ', end='')
 
     # Pretend to be an iterator when called as one and don't print any
     # progress
@@ -1762,6 +1845,17 @@ else:
     ProgressBar = ProgressBarTqdm
 
 
+class RealPathAction(argparse.Action):
+    def __init__(self, option_strings: T.List[str], dest: str, default: str = '.', **kwargs: T.Any):
+        default = os.path.abspath(os.path.realpath(default))
+        super().__init__(option_strings, dest, nargs=None, default=default, **kwargs)
+
+    def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace,
+                 values: T.Union[str, T.Sequence[T.Any], None], option_string: str = None) -> None:
+        assert isinstance(values, str)
+        setattr(namespace, self.dest, os.path.abspath(os.path.realpath(values)))
+
+
 def get_wine_shortpath(winecmd: T.List[str], wine_paths: T.Sequence[str]) -> str:
     """Get A short version of @wine_paths to avoid reaching WINEPATH number
     of char limit.
@@ -1770,11 +1864,11 @@ def get_wine_shortpath(winecmd: T.List[str], wine_paths: T.Sequence[str]) -> str
     wine_paths = list(OrderedSet(wine_paths))
 
     getShortPathScript = '%s.bat' % str(uuid.uuid4()).lower()[:5]
-    with open(getShortPathScript, mode='w') as f:
+    with open(getShortPathScript, mode='w', encoding='utf-8') as f:
         f.write("@ECHO OFF\nfor %%x in (%*) do (\n echo|set /p=;%~sx\n)\n")
         f.flush()
     try:
-        with open(os.devnull, 'w') as stderr:
+        with open(os.devnull, 'w', encoding='utf-8') as stderr:
             wine_path = subprocess.check_output(
                 winecmd +
                 ['cmd', '/C', getShortPathScript] + wine_paths,
@@ -1808,16 +1902,6 @@ def run_once(func: T.Callable[..., _T]) -> T.Callable[..., _T]:
     return wrapper
 
 
-class OptionProxy(T.Generic[_T]):
-    def __init__(self, value: _T, choices: T.Optional[T.List[str]] = None):
-        self.value = value
-        self.choices = choices
-
-    def set_value(self, v: _T) -> None:
-        # XXX: should this be an error
-        self.value = v
-
-
 class OptionOverrideProxy(collections.abc.MutableMapping):
 
     '''Mimic an option list but transparently override selected option
@@ -1833,15 +1917,16 @@ class OptionOverrideProxy(collections.abc.MutableMapping):
         for o in options:
             self.options.update(o)
 
-    def __getitem__(self, key: 'OptionKey') -> T.Union['UserOption', OptionProxy]:
+    def __getitem__(self, key: 'OptionKey') -> T.Union['UserOption']:
         if key in self.options:
             opt = self.options[key]
             if key in self.overrides:
-                return OptionProxy(opt.validate_value(self.overrides[key]), getattr(opt, 'choices', None))
+                opt = copy.copy(opt)
+                opt.set_value(self.overrides[key])
             return opt
         raise KeyError('Option not found', key)
 
-    def __setitem__(self, key: 'OptionKey', value: T.Union['UserOption', OptionProxy]) -> None:
+    def __setitem__(self, key: 'OptionKey', value: T.Union['UserOption']) -> None:
         self.overrides[key] = value.value
 
     def __delitem__(self, key: 'OptionKey') -> None:
@@ -1910,7 +1995,6 @@ def _classify_argument(key: 'OptionKey') -> OptionType:
     """Classify arguments into groups so we know which dict to assign them to."""
 
     if key.name.startswith('b_'):
-        assert key.machine is MachineChoice.HOST, str(key)
         return OptionType.BASE
     elif key.lang is not None:
         return OptionType.COMPILER
@@ -2023,14 +2107,13 @@ class OptionKey:
         This takes strings like `mysubproject:build.myoption` and Creates an
         OptionKey out of them.
         """
-
         try:
             subproject, raw2 = raw.split(':')
         except ValueError:
             subproject, raw2 = '', raw
 
         if raw2.startswith('build.'):
-            raw3 = raw2.lstrip('build.')
+            raw3 = raw2.split('.', 1)[1]
             for_machine = MachineChoice.BUILD
         else:
             raw3 = raw2
